from datetime import datetime
import json
import logging
from os.path import abspath, expanduser, isfile
from typing import List

import numpy as np
import pandas as pd
import requests

import geopandas as gpd

from ..dataframe_utils import merge_df, resample_df
from ..variables import MesowestVariables, SensorDescription
from .base import PointData

LOG = logging.getLogger("metloom.pointdata.mesowest")


class MesowestPointData(PointData):
    ALLOWED_VARIABLES = MesowestVariables
    MESO_URL = "https://api.synopticdata.com/v2/stations/timeseries"
    META_URL = "https://api.synopticdata.com/v2/stations/metadata"
    DATASOURCE = "Mesowest"
    POINTS_FROM_GEOM_DEFAULTS = {
        'within_geometry': True,
        'token_json': "~/.synoptic_token.json",
        'buffer': 0.0
    }
    NO_DATA_MESSAGE = 'No stations found for this request.'

    def __init__(self, station_id, name,
                 token_json="~/.synoptic_token.json", metadata=None):
        super(MesowestPointData, self).__init__(station_id, name, metadata=metadata)
        self._raw_metadata = None
        self._raw_elements = None
        self._tzinfo = None
        self._token = None
        self._token_json = token_json
        # Add the token to our urls
        self._meta_url = self.META_URL + f"?token={self.token}"
        self._meso_url = self.MESO_URL + f"?token={self.token}"

    @classmethod
    def get_token(cls, token_json):
        """
        Return the token stored in the json
        """
        token_json = abspath(expanduser(token_json))
        if not isfile(token_json):
            raise FileNotFoundError(f"Token file missing. Please sign up for a token "
                                    "with Synoptic Labs and add it to a json using "
                                    "MesowestPointData.create_token_json(token)\n "
                                    f"Missing {token_json}!")

        with open(token_json) as fp:
            token = json.load(fp)['token']
        return token

    @property
    def token(self):
        if self._token is None:
            self._token = self.get_token(self._token_json)
        return self._token

    def _get_metadata(self):
        """
        Method to get a shapely Point object to describe the station location
        which is assigned to the metadata property

        Returns:
            shapely.point.Point object in Longitude, Latitude
        """
        shp_point = None
        if self._raw_metadata is None:
            resp = requests.get(self._meta_url, params={"stid": self.id})
            resp.raise_for_status()
            jresp = resp.json()

            self._raw_metadata = jresp["STATION"][0]

            shp_point = gpd.points_from_xy(
                [float(self._raw_metadata["LONGITUDE"])],
                [float(self._raw_metadata["LATITUDE"])],
                z=[float(self._raw_metadata["ELEVATION"])],
            )[0]

        return shp_point

    def _get_data(self,
                  start_date: datetime,
                  end_date: datetime,
                  variables: List[SensorDescription],
                  interval='H'):
        """
        Make get request to Mesowest and return JSON
        Args:
            start_date: datetime object for start of data collection period
            end_date: datetime object for end of data collection period
            variables: List of metloom.variables.SensorDescription object
                from self.ALLOWED_VARIABLES
            interval: String interval the resulting data is resampled to
        Returns:
            dictionary of response values
        """
        fmt = '%Y%m%d%H%M'
        params = {
            "stid": self.id,
            "start": start_date.strftime(fmt),
            "end": end_date.strftime(fmt),
            "vars": ",".join([s.code for s in variables]),
            'units': 'metric',
            "obtimezone": "UTC",  # this is the default
        }
        resp = requests.get(self._meso_url, params=params)
        resp.raise_for_status()
        response_data = resp.json()

        final_columns = ["geometry", "site"]

        df = None
        for sensor in variables:
            if response_data:
                sensor_df = self._sensor_response_to_df(
                    response_data, sensor, final_columns, interval=interval)
                df = merge_df(df, sensor_df)

        if df is not None:
            df = df.dropna(axis=0)
            if len(df.index) > 0:
                # Set the datasource
                df["datasource"] = [self.DATASOURCE] * len(df.index)
                df.reset_index(inplace=True)
                df.set_index(keys=["datetime", "site"], inplace=True)
                df.index.set_names(["datetime", "site"], inplace=True)
            else:
                df = None
        self.validate_sensor_df(df)
        return df

    @staticmethod
    def _choose_sensor_key(timeseries_response, sensor):
        """
        Choose between possible keys for a variable

        Args:
            timeseries_response: the ["OBSERVATIONS"] return for a station
            sensor: the mesowest variable sensor description

        Returns:
            key for the data or None
        """
        # check that the variable was returned
        set1 = f"{sensor.code}_set_1"
        set1d = f"{sensor.code}_set_1d"
        # choose the best possible response
        if set1 not in timeseries_response and set1d not in timeseries_response:
            return None
        elif set1 in timeseries_response and set1d in timeseries_response:
            len_vals = len([val for val in timeseries_response[set1]
                            if val is not None and not np.isnan(val)])
            len_vals2 = len([val for val in
                            timeseries_response[set1d]
                             if val is not None and not np.isnan(val)])
            sensor_col = set1
            if len_vals < len_vals2:
                sensor_col = set1d
        elif set1 in timeseries_response:
            sensor_col = set1
        else:
            sensor_col = set1d
        return sensor_col

    def _sensor_response_to_df(self, response_data, sensor, final_columns,
                               interval: str = 'H'):
        """
        Convert the response data from the API to a GeoDataFrame
        Format and map columns in the dataframe
        Args:
            response_data: JSON list response from CDEC API
            sensor: SensorDescription obj
            final_columns: List of columns used for filtering
            interval: string interval for resampling
        Returns:
            GeoDataFrame
        """
        # handle no data returned
        response_summary = response_data.get('SUMMARY')
        if response_summary and response_summary.get(
            "RESPONSE_MESSAGE"
        ) == self.NO_DATA_MESSAGE:
            return None
        # parse the data for the station
        station_response = response_data['STATION']
        if len(station_response) == 0:
            return None
        else:
            timeseries_response = station_response[0]['OBSERVATIONS']
        # check that the variable was returned
        sensor_col = self._choose_sensor_key(timeseries_response, sensor)
        if sensor_col is None:
            return None

        sensor_df = pd.DataFrame.from_dict(
            timeseries_response
        )
        sensor_df.replace(-9999.0, np.nan, inplace=True)
        # This mapping is important.
        sensor_df.rename(
            columns={
                "date_time": "datetime",
                sensor_col: sensor.name,
            },
            inplace=True,
        )
        final_columns += [sensor.name, f"{sensor.name}_units"]

        sensor_df["datetime"] = pd.to_datetime(sensor_df["datetime"])
        # set index so joining works
        sensor_df.set_index("datetime", inplace=True)
        # handle timezones
        # mesowest can return utc so we do not need to localize
        sensor_df = sensor_df.tz_convert(self.desired_tzinfo)
        sensor_df = sensor_df.filter(final_columns)

        # make sure we resample to dataframe to the desired output freq
        sensor_df = resample_df(sensor_df, sensor, interval=interval)
        if sensor_df is None:
            return None
        # clean up dataframe and add other columns
        sensor_df = sensor_df.dropna(axis=0)
        sensor_df = gpd.GeoDataFrame(
            sensor_df,
            geometry=[self.metadata] * len(sensor_df),
        )
        sensor_df["site"] = [self.id] * len(sensor_df)

        sensor_df[f"{sensor.name}_units"] = response_data['UNITS'][sensor.code]

        return sensor_df

    def get_hourly_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        Get hourly measurement data
        Args:
            start_date: datetime object for start of data collection period
            end_date: datetime object for end of data collection period
            variables: List of metloom.variables.SensorDescription object
                from self.ALLOWED_VARIABLES
        Returns:
            GeoDataFrame of data. The dataframe should be indexed on
            ['datetime', 'site'] and have columns
            ['geometry', 'site']. Additionally, for each
            variables, it should have column f'{variable.name}' and
            f'{variable.name}_UNITS'
            See CDECPointData._get_data for example implementation and
            TestCDECStation.tny_daily_expected for example dataframe.
            Datetimes should be in UTC
        """
        df = self._get_data(start_date, end_date, variables, interval='H')
        return df

    def get_daily_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        Get daily measurement data
        Args:
            start_date: datetime object for start of data collection period
            end_date: datetime object for end of data collection period
            variables: List of metloom.variables.SensorDescription object
                from self.ALLOWED_VARIABLES
        Returns:
            GeoDataFrame of data. The dataframe should be indexed on
            ['datetime', 'site'] and have columns
            ['geometry', 'site']. Additionally, for each
            variables, it should have column f'{variable.name}' and
            f'{variable.name}_UNITS'
            See CDECPointData._get_data for example implementation and
            TestCDECStation.tny_daily_expected for example dataframe.
            Datetimes should be in UTC
        """
        df = self._get_data(start_date, end_date, variables, interval='D')
        return df

    @classmethod
    def points_from_geometry(
        cls,
        geometry: gpd.GeoDataFrame,
        variables: List[SensorDescription],
        **kwargs
    ):
        """
        See docstring for PointData.points_from_geometry

        Args:
            geometry: GeoDataFrame for shapefile from gpd.read_file
            variables: List of SensorDescription
            within_geometry: filter the points to within the shapefile
                instead of just the extents. Default True
            buffer: buffer added to search box
            token_json: Path to the public token for the mesowest api
                        default = "~/.synoptic_token.json"

        Returns:
            PointDataCollection
        """
        # assign defaults
        kwargs = cls._add_default_kwargs(kwargs)

        token = cls.get_token(kwargs['token_json'])
        projected_geom = geometry.to_crs(4326)
        bounds = projected_geom.bounds.iloc[0]
        buffer = kwargs["buffer"]
        adjusted_bounds = [
            bounds["minx"] - buffer, bounds["miny"] - buffer,
            bounds["maxx"] + buffer, bounds["maxy"] + buffer,
        ]
        bbox_str = ','.join(str(ab) for ab in adjusted_bounds)
        var_list_str = ','.join([v.code for v in variables])

        # Grab all the stations with the variables we want within a bounding box
        resp = requests.get(cls.META_URL + f"?token={token}",
                            params={'bbox': bbox_str, 'vars': var_list_str})
        resp.raise_for_status()

        points = []
        if resp:
            jdata = resp.json()
            data = jdata['STATION']

            points = []
            for sta in data:
                points.append(MesowestPointData(station_id=sta['STID'],
                                                name=sta['NAME'],
                                                token_json=kwargs['token_json']))
        # build the result geodataframe
        result_df = gpd.GeoDataFrame.from_dict({'STID': [p.id for p in points],
                                                'NAME': [p.name for p in points]},
                                               geometry=[p.metadata for p in points])

        # filter to points within shapefile
        if kwargs['within_geometry']:
            filtered_gdf = result_df[result_df.within(
                projected_geom.iloc[0]["geometry"])]
            points = [p for p in points if p.id in filtered_gdf['STID'].values]

        return cls.ITERATOR_CLASS(points)

    @classmethod
    def create_token_json(token):
        """
        Creates the neccessary synoptic token json for mesowest requests.
        To get public token visit: https://synopticdata.com/mesonet-api
        Args:
            token: Syntoptic Lab's public token.
        Returns:
            None
        """
        json_dict = {'token': token}
        with open(abspath(expanduser("~/.synoptic_token.json")), 'w') as outfile:
            json.dump(json_dict, outfile)



import json
import os
from collections import OrderedDict
from datetime import datetime
from os import path
from unittest.mock import MagicMock, patch

import pandas as pd
import pytest

import geopandas as gpd
from metloom.pointdata.mesowest import MesowestPointData
from metloom.variables import MesowestVariables
from tests.test_point_data import BasePointDataTest


class TestMesowestPointData(BasePointDataTest):
    @pytest.fixture(scope='session')
    def token_file(self):
        """
        Json token file fixture for mocking having a token
        """
        d = {'token': '####'}
        json_file = path.join(path.dirname(__file__), 'token.json')

        with open(json_file, 'w+') as fp:
            json.dump(d, fp)

        yield json_file
        # Clean up
        if path.isfile(json_file):
            os.remove(json_file)

    @pytest.fixture(scope="class")
    def data_dir(self):
        this_dir = path.dirname(__file__)
        return path.join(this_dir, "data")

    @pytest.fixture(scope="class")
    def shape_obj(self, data_dir):
        fp = path.join(data_dir, "triangle.shp")
        return gpd.read_file(fp)

    def _meta_response(self, *args, **kwargs):
        """
        Mccall airport station metadata return
        """
        mock = MagicMock()

        response = {}
        params = kwargs["params"]

        if 'stid' in params.keys():
            stid = params['stid']
            if stid == 'KMYL':
                response = {'STATION': [{'ELEVATION': '5020',
                                         'NAME': 'McCall Airport',
                                         'STID': 'KMYL',
                                         'ELEV_DEM': '5006.6',
                                         'LONGITUDE': '-116.09978',
                                         'LATITUDE': '44.89425',
                                         'TIMEZONE': 'America/Boise'}]}

            elif stid == 'INTRI':
                response = {'STATION': [{'ELEVATION': '9409',
                                         'NAME': 'IN TRIANGLE',
                                         'STID': 'INTRI',
                                         'LONGITUDE': '-119.5',
                                         'LATITUDE': '38.0',
                                         'TIMEZONE': 'America/Los_Angeles',
                                         }]}

            elif stid == 'OUTTRI':
                response = {'STATION': [
                    {'ELEVATION': '7201',
                     'NAME': 'OUT TRIANGLE W/IN BOUNDS',
                     'STID': 'OUTTRI',
                     'TIMEZONE': 'America/Los_Angeles',
                     'LONGITUDE': '-119.7',
                     'LATITUDE': '38.0',
                     }]}

        elif 'bbox' in params.keys():
            response = {'STATION': [{'ELEVATION': '9409',
                                     'NAME': 'IN TRIANGLE',
                                     'STID': 'INTRI',
                                     'LONGITUDE': '-119.5',
                                     'LATITUDE': '38.0',
                                     'TIMEZONE': 'America/Los_Angeles',
                                     },
                                    {'ELEVATION': '7201',
                                     'NAME': 'OUT TRIANGLE W/IN BOUNDS',
                                     'STID': 'OUTTRI',
                                     'TIMEZONE': 'America/Los_Angeles',
                                     'LONGITUDE': '-119.7',
                                     'LATITUDE': '38.0',
                                     }
                                    ]}
        else:
            raise ValueError('Invalid test STID provided')

        mock.json.return_value = response
        return mock

    @pytest.fixture(scope="class")
    def nodata_response(self):
        """
        Mesowest api return when no data is found for a variable for one
        station
        """
        mock = MagicMock()
        response = {
            "SUMMARY": {
                "RESPONSE_MESSAGE": 'No stations found for this request.'
            }
        }
        mock.json.return_value = response
        return mock

    @staticmethod
    def ts_response(var, values, delta, units: str):
        """
        Build a timeseries return using only values and assuming a date start
        point. We do this to avoid a lot of lists of repeat dates.
        """

        # Build a datetime list to match the values
        fmt = '%Y-%m-%dT%H:%M:%SZ'
        dt = [(pd.to_datetime('2021-01-01T00:00') + delta * i).strftime(fmt)
              for i in range(len(values))]

        # Populate the response
        response = {'UNITS': {var: units},
                    'STATION': [{
                        'NAME': 'McCall Airport',
                        'STID': 'KMYL',
                        'SENSOR_VARIABLES': {
                            'date_time': {'date_time': {},
                                          var: {f'{var}_set_1': {
                                              'position': '6.56'}}}},
                        'ELEVATION': '5006.6',
                        'LONGITUDE': '-119.5',
                        'LATITUDE': '44.89425',
                        'OBSERVATIONS': {
                            'date_time': dt,
                            f'{var}_set_1': values},
                        'TIMEZONE': 'UTC'}]}
        return response

    @pytest.fixture()
    def bbox_response(self):
        """
        Metadata response from mesowest using a bbox
        """
        mock = MagicMock()
        response = {'STATION': [{'ELEVATION': '9409',
                                 'NAME': 'IN TRIANGLE',
                                 'STID': 'INTRI',
                                 'LONGITUDE': '-119.5',
                                 'LATITUDE': '38.0',
                                 'TIMEZONE': 'America/Los_Angeles',
                                 },
                                {'ELEVATION': '7201',
                                 'NAME': 'OUT TRIANGLE W/IN BOUNDS',
                                 'STID': 'OUTTRI',
                                 'TIMEZONE': 'America/Los_Angeles',
                                 'LONGITUDE': '-119.7',
                                 'LATITUDE': '38.0',
                                 }
                                ]}
        mock.json.return_value = response
        return mock

    @pytest.fixture()
    def sub_hr_response(self, var, values, units):
        """
        Build a sub hourly response for time series data
        """
        mock = MagicMock()

        delta = pd.to_timedelta(30, 'minutes')
        response = self.ts_response(var.code, values, delta, units)
        mock.json.return_value = response
        return mock

    @pytest.fixture()
    def sub_daily_response(self, var, values, units):
        """
        Build a sub daily response for time series data
        """
        delta = pd.to_timedelta(12, 'hours')
        return self.ts_response(var.code, values, delta, units)

    @pytest.fixture()
    def station(self, token_file):
        return MesowestPointData("KMYL", "Mccall Airport", token_json=token_file)

    @pytest.mark.parametrize('stid, long, lat, elev', [
        ("KMYL", -116.09978, 44.89425, 5020.0),
    ])
    def test_get_metadata(self, token_file, stid, long, lat, elev):
        result = False
        with patch("metloom.pointdata.mesowest.requests") as mock_requests:
            mock_get = mock_requests.get
            mock_get.side_effect = self._meta_response
            expected = gpd.points_from_xy([long], [lat], z=[elev])[0]
            station = MesowestPointData(stid, 'test', token_json=token_file)
            result = station.metadata == expected
        assert result

    @pytest.mark.parametrize('var, values, units, expected_values, expected_dates', [
        (MesowestVariables.TEMP, [14.0, 16.0, 16.0, 18.0], 'Celsius', [15.0, 17.0],
         ['2021-01-1T00:00:00+00:00', '2021-01-1T01:00:00+00:00']),
    ])
    def test_get_hourly_data(self, station, sub_hr_response,
                             var, values, units, expected_values, expected_dates):
        # Patch in the made up response
        with patch("metloom.pointdata.mesowest.requests.get",
                   return_value=sub_hr_response):
            df = station.get_hourly_data(
                datetime(2021, 1, 1, 0),
                datetime(2021, 1, 1, 2),
                [var],
            )

        dt = [pd.to_datetime(d) for d in expected_dates]
        data = sub_hr_response.json()['STATION'][0]
        shp_point = gpd.points_from_xy(
            [float(data["LONGITUDE"])],
            [float(data["LATITUDE"])],
            z=[float(data["ELEVATION"])],
        )[0]

        expected = gpd.GeoDataFrame.from_dict(
            OrderedDict({
                'site': [station.id] * len(dt),
                var.name: expected_values,
                'geometry': [shp_point] * len(dt),
                'datetime': dt,
                f'{var.name}_units': [units] * len(dt),
                'datasource': ["Mesowest"] * len(dt),
            }),
            geometry=[shp_point] * len(dt))
        expected.set_index(keys=["datetime", "site"], inplace=True)
        pd.testing.assert_frame_equal(df, expected)

    def test_get_hourly_nodata(self, station, nodata_response):
        # Patch in the made up response
        with patch("metloom.pointdata.mesowest.requests.get",
                   return_value=nodata_response):
            df = station.get_hourly_data(
                datetime(2021, 1, 1, 0),
                datetime(2021, 1, 1, 2),
                [MesowestVariables.TEMP]
            )
        assert df is None

    @pytest.mark.parametrize('w_geom, expected_sid', [
        (False, ['INTRI', 'OUTTRI']),  # Use just bounds of the shapefile
        (True, ['INTRI']),  # Filter to within the shapefile
    ])
    def test_points_from_geometry(self, token_file, shape_obj, w_geom, expected_sid):

        with patch("metloom.pointdata.mesowest.requests") as mock_requests:
            mock_get = mock_requests.get
            mock_get.side_effect = self._meta_response
            pnts = MesowestPointData.points_from_geometry(shape_obj,
                                                          [MesowestVariables.TEMP],
                                                          within_geometry=w_geom,
                                                          token_json=token_file)

        df = pnts.to_dataframe()
        assert df['id'].values == pytest.approx(expected_sid)

    def test_points_from_geometry_buffer(self, token_file, shape_obj):

        with patch("metloom.pointdata.mesowest.requests") as mock_requests:
            mock_get = mock_requests.get
            mock_get.side_effect = self._meta_response
            MesowestPointData.points_from_geometry(
                shape_obj, [MesowestVariables.TEMP],
                within_geometry=False, token_json=token_file,
                buffer=0.1
            )
            call_params = mock_get.call_args_list[0][1]["params"]

        results = [float(v) for v in call_params["bbox"].split(',')]
        expected = [-119.9, 37.6, -119.1, 38.3]
        for result, exp in zip(results, expected):
            assert exp == pytest.approx(result)

    def test_missing_token_instantiation(self):
        """
        Test the missing token file raises an IOerror when instantiated
        """
        with pytest.raises(FileNotFoundError):
            MesowestPointData('TEST', 'test',
                              token_json="./non-existent_path.json")

    def test_missing_token_class_method(self, shape_obj):
        """
        Test the missing token file raises an IOerror when data is requested vi
        a class method
        """
        with patch("metloom.pointdata.mesowest.requests") as mock_requests:
            mock_get = mock_requests.get
            mock_get.side_effect = self._meta_response
            with pytest.raises(FileNotFoundError):
                MesowestPointData.points_from_geometry(
                    shape_obj,
                    [MesowestVariables.TEMP],
                    token_json="./non-existent_path.json")

    @pytest.mark.parametrize(
        "timeseries, expected", [
            ({
                'date_time': [datetime(2020, 1, 1), datetime(2020, 1, 2)],
                'pressure_set_1': [10.1, 10.2],
                'pressure_set_1d': [10.2, 10.3]
            }, "pressure_set_1"),
            ({
                'date_time': [datetime(2020, 1, 1), datetime(2020, 1, 2)],
                'pressure_set_1': [10.1, 10.2],
                'pressure_set_1d': [10.2, None]
            }, "pressure_set_1"),
            ({
                'date_time': [datetime(2020, 1, 1), datetime(2020, 1, 2)],
                'pressure_set_1': [10.1, None],
                'pressure_set_1d': [10.2, 10.3]
            }, "pressure_set_1d"),
            ({
                'date_time': [datetime(2020, 1, 1), datetime(2020, 1, 2)],
                'temperature_set_1': [5.2, 5.3]
            }, None)
        ]
    )
    def test_choose_sensor_key(self, timeseries, expected):
        result = MesowestPointData._choose_sensor_key(
            timeseries, MesowestVariables.PRESSURE
        )
        assert result == expected



import pytest

from metloom.variables import VariableBase, CdecStationVariables


class TestBaseVariables:
    def test_validate_variable_fails(self):
        with pytest.raises(ValueError):
            VariableBase.from_code(-1)


class TestCDECStationVariables:
    @pytest.mark.parametrize(
        "code, expected",
        [
            (2, CdecStationVariables.PRECIPITATIONACCUM),
            ("2", CdecStationVariables.PRECIPITATIONACCUM),
            ("3", CdecStationVariables.SWE),
            (30, CdecStationVariables.TEMPAVG),
        ],
    )
    def test_from_code(self, code, expected):
        assert CdecStationVariables.from_code(code) == expected

    def test_from_code_failure(self):
        with pytest.raises(ValueError):
            CdecStationVariables.from_code(-9999)



from .base import PointData, PointDataCollection
from .cdec import CDECPointData
from .snotel import SnotelPointData
from .mesowest import MesowestPointData
from .usgs import USGSPointData

__all__ = [
    "PointData", "PointDataCollection", "CDECPointData", "SnotelPointData",
    "MesowestPointData", "USGSPointData"
]



from dataclasses import dataclass


@dataclass(eq=True, frozen=True)
class SensorDescription:
    """
    data class for describing a snow sensor
    """

    code: str = "-1"  # code used within the applicable API
    name: str = "basename"  # desired name for the sensor
    description: str = None  # description of the sensor
    accumulated: bool = False  # whether or not the data is accumulated


class VariableBase:
    """
    Base class to store all variables for a specific datasource. Each
    datasource should implement the class. The goal is that the variables
    are synonymous across implementations.(i.e. PRECIPITATION should have the
    same meaning in each implementation).
    Additionally, variables with the same meaning should have the same
    `name` attribute of the SensorDescription. This way, if multiple datsources
    are used to sample the same variable, they can be written to the same
    column in a csv.

    Variables in this base class should ideally be implemented by all classes
    and cannot be directly used from the base class.
    """

    PRECIPITATION = SensorDescription()
    SWE = SensorDescription()
    SNOWDEPTH = SensorDescription()

    @staticmethod
    def _validate_sensor(sensor: SensorDescription):
        """
        Validate that a sensor is not using the default values since they
        are meaningless
        """
        default = SensorDescription()
        if sensor.name == default.name and sensor.code == default.code:
            raise ValueError(f"{sensor.name} is the default implementation")

    @classmethod
    def from_code(cls, code):
        """
        Get the correct sensor description from the code
        """
        for k, v in cls.__dict__.items():
            if isinstance(v, SensorDescription) and v.code == str(code):
                cls._validate_sensor(v)
                return v
        raise ValueError(f"Could not find sensor for code {code}")


class CdecStationVariables(VariableBase):
    """
    Available sensors from CDEC.
    Exhaustive list:
    http://cdec4gov.water.ca.gov/reportapp/javareports?name=SensList
    """

    PRECIPITATIONACCUM = SensorDescription(
        "2", "ACCUMULATED PRECIPITATION", "PRECIPITATION, ACCUMULATED", False
    )
    PRECIPITATION = SensorDescription(
        "45", "PRECIPITATION", "PRECIPITATION, INCREMENTAL", True
    )
    SNOWDEPTH = SensorDescription("18", "SNOWDEPTH", "SNOW DEPTH")
    SWE = SensorDescription("3", "SWE", "SNOW, WATER CONTENT", False)
    TEMP = SensorDescription("4", "AIR TEMP", "TEMPERATURE, AIR")
    TEMPAVG = SensorDescription("30", "AVG AIR TEMP", "TEMPERATURE, AIR AVERAGE")
    TEMPMIN = SensorDescription("32", "MIN AIR TEMP", "TEMPERATURE, AIR MINIMUM")
    TEMPMAX = SensorDescription("31", "MAX AIR TEMP", "TEMPERATURE, AIR MAXIMUM")
    RH = SensorDescription("12", "Relative Humidity", "RELATIVE HUMIDITY")
    # TODO confirm with CDWR if these depths are standard, no metadata available
    TEMPGROUND = SensorDescription(
        "52", "GROUND TEMPERATURE INT", "GROUND TEMPERATURE SNOW/SOIL INTERFACE"
    )
    TEMPGROUND25CM = SensorDescription(
        "194", "GROUND TEMPERATURE -25CM", "GROUND TEMPERATURE OBS -25CM"
    )
    TEMPGROUND50CM = SensorDescription(
        "195", "GROUND TEMPERATURE -50CM", "GROUND TEMPERATURE OBS -50CM"
    )
    TEMPGROUND100CM = SensorDescription(
        "196", "GROUND TEMPERATURE -100CM", "GROUND TEMPERATURE OBS -100CM"
    )
    SOLARRAD = SensorDescription("103", "SOLAR RADIATION", "SOLAR RADIATION")
    WINDSPEED = SensorDescription("9", "WIND SPEED", "WIND SPEED")
    WINDDIR = SensorDescription("10", "WIND DIRECTION", "WIND DIRECTION")


class SnotelVariables(VariableBase):
    """
    Available sensors from SNOTEL
    """

    SNOWDEPTH = SensorDescription("SNWD", "SNOWDEPTH")
    SWE = SensorDescription("WTEQ", "SWE")
    TEMP = SensorDescription("TOBS", "AIR TEMP")
    TEMPAVG = SensorDescription("TAVG", "AVG AIR TEMP", "AIR TEMPERATURE AVERAGE")
    TEMPMIN = SensorDescription("TMIN", "MIN AIR TEMP", "AIR TEMPERATURE MINIMUM")
    TEMPMAX = SensorDescription("TMAX", "MAX AIR TEMP", "AIR TEMPERATURE MAXIMUM")
    PRECIPITATION = SensorDescription(
        "PRCPSA", "PRECIPITATION", "PRECIPITATION INCREMENT SNOW-ADJUSTED"
    )
    PRECIPITATIONACCUM = SensorDescription(
        "PREC", "ACCUMULATED PRECIPITATION", "PRECIPITATION ACCUMULATION"
    )
    TEMPGROUND2IN = SensorDescription(
        "STO", "GROUND TEMPERATURE -2IN", "GROUND TEMPERATURE OBS -2IN"
    )
    TEMPGROUND4IN = SensorDescription(
        "STO", "GROUND TEMPERATURE -4IN", "GROUND TEMPERATURE OBS -4IN"
    )
    TEMPGROUND8IN = SensorDescription(
        "STO", "GROUND TEMPERATURE -8IN", "GROUND TEMPERATURE OBS -8IN"
    )
    TEMPGROUND20IN = SensorDescription(
        "STO", "GROUND TEMPERATURE -20IN", "GROUND TEMPERATURE OBS -20IN"
    )
    SOILMOISTURE2IN = SensorDescription(
        "SMS", "SOIL MOISTURE -2IN", "SOIL MOISTURE PERCENT -2IN"
    )
    SOILMOISTURE4IN = SensorDescription(
        "SMS", "SOIL MOISTURE -4IN", "SOIL MOISTURE PERCENT -4IN"
    )
    SOILMOISTURE8IN = SensorDescription(
        "SMS", "SOIL MOISTURE -8IN", "SOIL MOISTURE PERCENT -8IN"
    )
    SOILMOISTURE20IN = SensorDescription(
        "SMS", "SOIL MOISTURE -20IN", "SOIL MOISTURE PERCENT -20IN"
    )
    # TODO for the SCAN network this appears to be "RHUM", we may need a new class
    RH = SensorDescription("RHUMV", "RELATIVE HUMIDITY", "RELATIVE HUMIDITY")
    STREAMVOLUMEOBS = SensorDescription(
        "SRVO", "STREAM VOLUME OBS", "STREAM VOLUME OBS"
    )
    STREAMVOLUMEADJ = SensorDescription(
        "SRVOX", "STREAM VOLUME ADJ", "STREAM VOLUME ADJ"
    )


class MesowestVariables(VariableBase):
    """
    Available sensors from Mesowest
    There are a lot of variables here. Feel free to PR to add some
    https://developers.synopticdata.com/mesonet/v2/api-variables/
    """

    TEMP = SensorDescription("air_temp", "AIR TEMP")
    DEWPOINT = SensorDescription("dew_point_temperature", "DEW POINT TEMPERATURE")
    RH = SensorDescription("relative_humidity", "RELATIVE HUMIDITY")
    WINDSPEED = SensorDescription("wind_speed", "WIND SPEED")
    WINDDIRECTION = SensorDescription("wind_direction", "WIND DIRECTION")
    PRESSURE = SensorDescription("pressure", "PRESSURE")
    SNOWDEPTH = SensorDescription("snow_depth", "SNOWDEPTH")
    SOLARRADIATION = SensorDescription("solar_radiation", "SOLAR RADIATION")
    WETBULBTEMPERATURE = SensorDescription(
        "wet_bulb_temperature", "WET BULB TEMPERATURE"
    )
    SOILTEMP = SensorDescription("soil_temp", "SOIL TEMPERATURE")
    SOILTEMPIR = SensorDescription("soil_temp_ir", "SOIL TEMPERATURE IR")
    SWE = SensorDescription("snow_water_equiv", "SWE")
    NETSHORTWAVE = SensorDescription("net_radiation_sw", "NET SHORTWAVE RADIATION")
    NETLONGWAVE = SensorDescription("net_radiation_lw", "NET LONGWAVE RADIATION")
    STREAMFLOW = SensorDescription("stream_flow", "STREAMFLOW")


class USGSVariables(VariableBase):
    """
    To add more sensors:
    https://help.waterdata.usgs.gov/codes-and-parameters/parameters
    """
    DISCHARGE = SensorDescription("00060", "DISCHARGE", "DISCHARGE (CFS)")
    STREAMFLOW = SensorDescription(
        "74082", "STREAMFLOW", "STREAMFLOW, DAILY VOLUME (AC-FT)"
    )
    SNOWDEPTH = SensorDescription("72189", "SNOWDEPTH", "Snow depth, Meters")
    SWE = SensorDescription(
        "72341", "SWE", "Water content of snow, millimeters"
    )



import numpy as np
import pandas as pd
import pytest
from datetime import datetime, timedelta
from metloom.variables import MesowestVariables, CdecStationVariables

from metloom.dataframe_utils import (
    join_df, append_df, merge_df, resample_df, resample_whole_df
)

df1 = pd.DataFrame.from_records([{"foo": 12.0}, {"foo": 11.0}])
df2 = pd.DataFrame.from_records([{"bar": 1.0}, {"bar": 1.0}])
df3 = pd.DataFrame.from_records([
    {"foo": 12.0, "bar": 1.0}, {"foo": 11.0, "bar": 1.0}
])
df4 = pd.DataFrame.from_records(
    [
        {"foo": 12.0},
        {"foo": 11.0},
        {"foo": 12.0},
        {"foo": 11.0},
    ],
    index=[0, 1, 0, 1],
)


@pytest.mark.parametrize(
    "set1, set2, expected", [(df1, df2, df3), (df1, None, df1), (None, df2, df2)]
)
def test_join_df(set1, set2, expected):
    pd.testing.assert_frame_equal(join_df(set1, set2), expected)


def test_join_df_failure():
    with pytest.raises((ValueError, AttributeError)):
        join_df(df1, "bad value")


@pytest.mark.parametrize(
    "set1, set2, expected", [(df1, None, df1), (None, df2, df2), (df1, df1, df4)]
)
def test_append_df(set1, set2, expected):
    pd.testing.assert_frame_equal(append_df(set1, set2), expected)


def test_merge_df():
    first = pd.DataFrame(
        {
            "datetime": [
                datetime(2020, 1, 3), datetime(2020, 1, 6),
                datetime(2020, 1, 7)
            ],
            "a": [3.0, 6.0, 7.0],
            "c": [3.0, 6.0, 7.0]
        }
    ).set_index("datetime")
    second = pd.DataFrame(
        {
            "datetime": [
                datetime(2020, 1, 2), datetime(2020, 1, 5),
                datetime(2020, 1, 6), datetime(2020, 1, 8)],
            "b": [2.0, 5.0, 6.0, 8.0],
            "c": [2.0, 5.0, 6.0, 8.0]
        },
    ).set_index("datetime")

    expected = pd.DataFrame(
        {
            "datetime": [
                datetime(2020, 1, 2), datetime(2020, 1, 3), datetime(2020, 1, 5),
                datetime(2020, 1, 6), datetime(2020, 1, 7), datetime(2020, 1, 8)],
            "a": [np.nan, 3.0, np.nan, 6.0, 7.0, np.nan],
            "c": [2.0, 3.0, 5.0, 6.0, 7.0, 8.0],
            "b": [2.0, np.nan, 5.0, 6.0, np.nan, 8.0],
        }
    ).set_index("datetime")
    result = merge_df(first, second)
    pd.testing.assert_frame_equal(expected, result)


def test_merge_df_duplicates():
    first = pd.DataFrame(
        {
            "datetime": [
                datetime(2020, 1, 2), datetime(2020, 1, 5)
            ],
            "b": [6.0, 8.0],
            "c": [6.0, 8.0],
        },
    ).set_index("datetime")
    second = pd.DataFrame(
        {
            "datetime": [
                datetime(2020, 1, 3), datetime(2020, 1, 3),
                datetime(2020, 1, 7)
            ],
            "a": [3.0, 3.0, 3.0],
            "c": [4.0, 4.0, 4.0]
        }
    ).set_index("datetime")
    result = merge_df(first, second)
    expected = pd.DataFrame(
        {
            "datetime": [
                datetime(2020, 1, 2),
                datetime(2020, 1, 3),
                datetime(2020, 1, 5), datetime(2020, 1, 7)
            ],
            "b": [6.0, np.nan, 8.0, np.nan],
            "c": [6.0, 4.0, 8.0, 4.0],
            "a": [np.nan, 3.0, np.nan, 3.0],
        }
    ).set_index("datetime")
    pd.testing.assert_frame_equal(result, expected)


@pytest.fixture()
def sample_df(in_data, variable, delta_t, interval):
    """
    Build a dataframe given a incoming data set and interval to
    look at and a delta time to create the datetime index.
    """
    if interval == 'H':
        dt = timedelta(minutes=delta_t)
    elif interval == 'D':
        dt = timedelta(hours=delta_t)

    d_time = [datetime(2021, 12, 1) + i * dt for i in range(len(in_data))]
    df = pd.DataFrame.from_dict({'datetime': d_time, variable.name: in_data})
    df = df.set_index('datetime')
    yield df


@pytest.mark.parametrize("in_data, variable, delta_t, interval, expected_data", [
    ([1, 3, 2, 4], MesowestVariables.TEMP, 30, 'H', [2, 3]),
    ([3, 3, 2, 2], CdecStationVariables.SWE, 12, 'D', [3, 2]),
])
def test_resample_df(sample_df, in_data, variable, delta_t, interval, expected_data):
    """
    Test the resample function can resample values according to
    hourly and daily time intervals
    """
    out_df = resample_df(sample_df, variable, interval=interval)
    assert out_df[variable.name].values == pytest.approx(expected_data)


def test_resample_df_odd_increment():
    df = pd.DataFrame.from_records([
        {"datetime": datetime(2020, 1, 2, 11, 3, 20), "AIR TEMP": 1.0},
        {"datetime": datetime(2020, 1, 2, 11, 13, 21), "AIR TEMP": 2.0},
        {"datetime": datetime(2020, 1, 2, 11, 42, 1), "AIR TEMP": 3.0},
        {"datetime": datetime(2020, 1, 2, 12, 5, 28), "AIR TEMP": 2.0},
        {"datetime": datetime(2020, 1, 2, 12, 59, 3), "AIR TEMP": 1.5},
    ])
    df.set_index("datetime", inplace=True)
    out_df = resample_df(df, MesowestVariables.TEMP, interval="H")
    expected = pd.DataFrame.from_records([
        {"datetime": datetime(2020, 1, 2, 11), "AIR TEMP": 2.0},
        {"datetime": datetime(2020, 1, 2, 12,), "AIR TEMP": 1.75},
    ])
    expected.set_index("datetime", inplace=True)
    expected.index.freq = "H"
    pd.testing.assert_frame_equal(out_df, expected)


def test_resample_full_df_odd_increment():
    df = pd.DataFrame.from_records([
        {"datetime": datetime(2020, 1, 2, 11, 3, 20), "other": "value",
         "AIR TEMP": 1.0},
        {"datetime": datetime(2020, 1, 2, 11, 13, 21), "other": "value",
         "AIR TEMP": 2.0},
        {"datetime": datetime(2020, 1, 2, 11, 42, 1), "other": "value",
         "AIR TEMP": 3.0},
        {"datetime": datetime(2020, 1, 2, 12, 5, 28), "other": "value",
         "AIR TEMP": 2.0},
        {"datetime": datetime(2020, 1, 2, 12, 59, 3), "other": "value",
         "AIR TEMP": 1.5},
    ])
    df.set_index("datetime", inplace=True)
    out_df = resample_whole_df(df, MesowestVariables.TEMP, interval="H")
    expected = pd.DataFrame.from_records([
        {"datetime": datetime(2020, 1, 2, 11), "other": "value", "AIR TEMP": 2.0},
        {"datetime": datetime(2020, 1, 2, 12,), "other": "value", "AIR TEMP": 1.75},
    ])
    expected.set_index("datetime", inplace=True)
    expected.index.freq = "H"
    pd.testing.assert_frame_equal(out_df, expected, check_like=True)



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'geopandas>=0.9.0,<1.0.0',
    'pandas>1.0.0',
    'lxml==4.9.3',
    'requests>2.0.0,<3.0.0',
    'beautifulsoup4>4,<5',
    'zeep>4.0.0'
]

test_requirements = ['pytest>=3', ]

setup(
    author="M3Works",
    author_email='m3worksllc@gmail.com',
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description="Location Oriented Observed Meteorology (LOOM)",
    entry_points={
        'console_scripts': [
            'metloom=metloom.cli:main'
        ],
    },
    install_requires=requirements,
    license="BSD license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='metloom',
    name='metloom',
    packages=find_packages(include=['metloom', 'metloom.*']),
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/M3Works/metloom',
    version='0.3.10',
    zip_safe=False,
)



from datetime import datetime
from os.path import join
from pathlib import Path
from unittest.mock import patch
import json

import geopandas as gpd
import pandas as pd
import pytest

from metloom.pointdata import USGSPointData
from metloom.variables import USGSVariables
from tests.test_point_data import BasePointDataTest

DATA_DIR = str(Path(__file__).parent.joinpath("data/usgs_mocks"))


class TestUSGSStation(BasePointDataTest):

    @staticmethod
    def station_search_response():
        with open(join(DATA_DIR, "station_search_response.txt")) as fp:
            data_text = fp.read()

        return data_text

    @pytest.fixture(scope="function")
    def crp_station(self):
        return USGSPointData("08245000", "Conejos R bl Platoro Reservoir")

    @pytest.fixture(scope="class")
    def crp_daily_expected(self):
        points = gpd.points_from_xy([-106.54], [37.35], z=[9866.6])
        df = gpd.GeoDataFrame.from_dict(
            [
                {
                    "datetime": pd.Timestamp("2020-07-01 07:00:00+0000", tz="UTC"),
                    "DISCHARGE": 721.0,
                    "DISCHARGE_units": "ft3/s",
                    "site": "08245000",
                    "datasource": "USGS",
                },
                {
                    "datetime": pd.Timestamp("2020-07-02 07:00:00+0000", tz="UTC"),
                    "DISCHARGE": 664.0,
                    "DISCHARGE_units": "ft3/s",
                    "site": "08245000",
                    "datasource": "USGS",
                },

            ],
            geometry=[points[0]] * 2,
        )
        # needed to reorder the columns for the pd testing compare
        df = df.filter(
            [
                "datetime",
                "geometry",
                "DISCHARGE",
                "site",
                "DISCHARGE_units",
                "datasource",
            ]
        )
        df.set_index(keys=["datetime", "site"], inplace=True)
        return df

    @staticmethod
    def failure_response():
        class Response:
            status_code = 404
            content = (
                b'<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" '
                b'"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd"><html xmlns='
                b'"http://www.w3.org/1999/xhtml"><head><title>Error report</title>'
                b'<style type="text/css"><!--H1 {font-family:Tahoma,Arial,sans-serif;'
                b'color:white;background-color:#525D76;font-size:22px;} H2 {font-family'
                b':Tahoma,Arial,sans-serif;color:white;background-color:#525D76;font-'
                b'size:16px;} H3 {font-family:Tahoma,Arial,sans-serif;color:white;'
                b'background-color:#525D76;font-size:14px;} BODY {font-family:Tahoma,'
                b'Arial,sans-serif;color:black;background-color:white;} B {font-family:'
                b'Tahoma,Arial,sans-serif;color:white;background-color:#525D76;} P '
                b'{font-family:Tahoma,Arial,sans-serif;background:white;color:black;'
                b'font-size:12px;}A {color : black;}HR {color : #525D76;}--></style> '
                b'</head><body><h1>HTTP Status 404 - No sites found matching this '
                b'request, server=[sdas01]</h1><hr/><p><b>type</b> Status report</p>'
                b'<p><b>message</b>No sites found matching this request, server='
                b'[sdas01]</p><p><b>description</b>The requested resource is not '
                b'available.</p><hr/><h3>Error Report</h3></body></html>')

        return Response

    @classmethod
    def get_url_response(cls, resp="daily"):
        if resp == 'daily':
            with open(join(DATA_DIR, "daily_response.txt")) as fp:
                data = json.load(fp)
        elif resp == 'metadata':
            with open(join(DATA_DIR, "platoro_meta.txt")) as fp:
                data = fp.read()
        elif resp == 'hourly':
            with open(join(DATA_DIR, "hourly_response.json")) as fp:
                data = json.load(fp)
        else:
            raise RuntimeError(f"{resp} is an unknown option")

        return data

    def test_get_metadata(self, crp_station):
        with patch("metloom.pointdata.usgs.USGSPointData._get_url_response") \
                as mock_request:
            mock_request.return_value = self.get_url_response(resp="metadata")
            metadata = crp_station.metadata

        expected = gpd.points_from_xy([-106.54], [37.35], z=[9866.6])[0]
        assert expected == metadata

    def test_get_daily_data(self, crp_station, crp_daily_expected):
        with patch("metloom.pointdata.usgs.USGSPointData._get_url_response") \
                as mock_requests:
            mock_requests.side_effect = [
                self.get_url_response(),
                self.get_url_response(resp='metadata')
            ]
            response = crp_station.get_daily_data(
                datetime(2020, 7, 1),
                datetime(2020, 7, 2),
                [USGSVariables.DISCHARGE],
            )
        pd.testing.assert_frame_equal(response, crp_daily_expected)

    def test_get_hourly_data(self, crp_station, crp_daily_expected):
        """
        Test that we resample from 15m to 1 hour correctly
        """
        with patch("metloom.pointdata.usgs.USGSPointData._get_url_response") \
                as mock_requests:
            mock_requests.side_effect = [
                self.get_url_response(resp='hourly'),
                self.get_url_response(resp='metadata')
            ]
            response = crp_station.get_hourly_data(
                datetime(2023, 1, 13),
                datetime(2023, 1, 13),
                [USGSVariables.DISCHARGE],
            )
        response = response.reset_index()
        assert response["datetime"].values[0] == pd.to_datetime("2023-01-13 07")
        assert response["datetime"].values[-1] == pd.to_datetime("2023-01-14 06")
        assert response["DISCHARGE"].values[0] == 300.0
        assert response["DISCHARGE"].values[-1] == 283.25
        assert all(response["site"].values == "08245000")

    def test_points_from_geometry(self, shape_obj):
        expected_url = (
            'https://waterservices.usgs.gov/nwis/site/?format=rdb&bBox=-119.8%2C37.7'
            '%2C-119.2%2C38.2&siteStatus=active&hasDataTypeCd=dv,iv&parameterCd=00060'
        )
        names = [
            'TUOLUMNER NR HETCH HETCHY CA',
            'TUOLUMNE R A GRAND CYN OF TUOLUMNE AB HETCH HETCHY',
            'MILL C BL LUNDY LK NR LEE VINING CA',
            'LEE VINING C BL SADDLEBAG LK NR LEE VINING CA'
        ]
        with patch("metloom.pointdata.usgs.USGSPointData._get_url_response") as mock_tb:
            mock_tb.return_value = self.station_search_response()
            result = USGSPointData.points_from_geometry(
                shape_obj, [USGSVariables.DISCHARGE]
            )
            assert mock_tb.call_args[0][0] == expected_url
            assert len(result) == 10
            assert [x.name in names for x in result.points]

    def test_points_from_geometry_failure(self, shape_obj):
        expected_url = (
            'https://waterservices.usgs.gov/nwis/site/?format=rdb&bBox=-119.8%2C37.7'
            '%2C-119.2%2C38.2&siteStatus=active&hasDataTypeCd=dv,iv&parameterCd=74082'
        )
        with patch("metloom.pointdata.usgs.requests.get") as mock_request:
            mock_request.return_value = self.failure_response()
            result = USGSPointData.points_from_geometry(
                shape_obj, [USGSVariables.STREAMFLOW]
            )
            assert result.points == []
            assert mock_request.call_args[0][0] == expected_url

    def test_geometry_failure_message(self, caplog, shape_obj):
        expected_error_msg = (
            "No data: HTTP Status 404 - No sites found matching this request, "
            "server=[sdas01]"
        )
        with patch("metloom.pointdata.usgs.requests.get") as mock_request:
            mock_request.return_value = self.failure_response()
            point = USGSPointData("123", "test")
            result = point._get_url_response("test")
            assert result == []

        assert expected_error_msg in caplog.text

    def test_check_dates(self, crp_station):
        start_date = datetime(2020, 7, 1, 1, 1, 1)
        end_date = datetime(2020, 7, 2, 1, 1, 1)
        checked_start, checked_end = crp_station._check_dates(
            start_date, end_date
        )

        assert start_date.date() == checked_start
        assert end_date.date() == checked_end

    def test_dates_fail(self, crp_station, caplog):
        start_date = datetime(2020, 7, 2, 1, 1, 1)
        end_date = datetime(2020, 7, 1, 1, 1, 1)
        error_msg = (
            f" end_date '{end_date.date()}' must be later than "
            f"start_date '{start_date.date()}'"
        )
        with pytest.raises(ValueError):
            checked_start, checked_end = crp_station._check_dates(
                start_date, end_date
            )

        assert error_msg in caplog.text
